/*
This example reproduces the steady-state results of Figure 6
in the following reference:
J. Palacin, M. Salleras, J. Samitier, S. Marco,
"Dynamic compact thermal models with multiple power sources:
application to an ultrathin chip stacking technology",
IEEE Transactions on Advanced Packaging, Vol. 28, No. 4, November 2005.
*/

#include <stdio.h>
#include "stimul.h"


#define NX 100
#define NY 100
#define AREA_CHIP (0.25e-4)
#define AREA_SOURCE (0.14e-4)
#define CHIP_LX (2*2821e-6)
#define CHIP_LY (AREA_CHIP/CHIP_LX)
#define SOURCE_LX (2*2129e-6) 
#define SOURCE_LY (AREA_SOURCE/SOURCE_LX)
#define KSI 117.5
#define KBCB 0.18
#define POWDENS (1./(AREA_SOURCE/2))
#define HEATCOEFF 2.


int main()
{
  int i1,j1,i2,j2;
  stml_grid *gq = stml_grid_new(NX,NY);
  stml_grid *gt = stml_grid_new(NX,NY);
  stml_grid *g0 = stml_grid_new(NX,NY);
  stml_multilayer *p = stml_multilayer_new(CHIP_LX,CHIP_LY,NX,NY);

  stml_grid_set(g0,0);

  i1 = NX * 0.5 * (CHIP_LX-SOURCE_LX)/CHIP_LX;
  j1 = NY * 0.5 * (CHIP_LY-SOURCE_LY)/CHIP_LY;
  i2 = NX-1 - NX * 0.5 * (CHIP_LX-SOURCE_LX)/CHIP_LX;
  j2 = NY-1 - NY * 0.5 * (CHIP_LY-SOURCE_LY)/CHIP_LY;
  stml_grid_rectangle(gq,i1,j1,i2,j2,POWDENS);

  stml_multilayer_isotherm(p,0,"ambient1");
  stml_multilayer_interface(p,HEATCOEFF);
  stml_multilayer_slab(p,10e-6,KBCB);
  stml_multilayer_sources(p,NULL,"chip1");
  stml_multilayer_slab(p,15e-6,KSI);
  stml_multilayer_slab(p,10e-6,KBCB);
  stml_multilayer_sources(p,NULL,"chip2");
  stml_multilayer_slab(p,15e-6,KSI);
  stml_multilayer_slab(p,10e-6,KBCB);
  stml_multilayer_sources(p,NULL,"chip3");
  stml_multilayer_slab(p,15e-6,KSI);
  stml_multilayer_slab(p,3e-6,KBCB);
  stml_multilayer_slab(p,500e-6,KSI);
  stml_multilayer_isotherm(p,0,"ambient2");
  stml_multilayer_finalize(p);


  stml_multilayer_modify(p,"chip1",gq);
  stml_multilayer_solve(p,"chip1",gt);
  printf("T11(inf)= %5.2f  ",stml_grid_read(gt,NX/2,NY/2));
  stml_multilayer_solve(p,"chip2",gt);
  printf("T12(inf)= %5.2f  ",stml_grid_read(gt,NX/2,NY/2));
  stml_multilayer_solve(p,"chip3",gt);
  printf("T13(inf)= %5.2f\n",stml_grid_read(gt,NX/2,NY/2));

  stml_multilayer_modify(p,"chip1",g0);
  stml_multilayer_modify(p,"chip2",gq);
  stml_multilayer_solve(p,"chip1",gt);
  printf("T21(inf)= %5.2f  ",stml_grid_read(gt,NX/2,NY/2));
  stml_multilayer_solve(p,"chip2",gt);
  printf("T22(inf)= %5.2f  ",stml_grid_read(gt,NX/2,NY/2));
  stml_multilayer_solve(p,"chip3",gt);
  printf("T23(inf)= %5.2f\n",stml_grid_read(gt,NX/2,NY/2));

  stml_multilayer_modify(p,"chip2",g0);
  stml_multilayer_modify(p,"chip3",gq);
  stml_multilayer_solve(p,"chip1",gt);
  printf("T31(inf)= %5.2f  ",stml_grid_read(gt,NX/2,NY/2));
  stml_multilayer_solve(p,"chip2",gt);
  printf("T32(inf)= %5.2f  ",stml_grid_read(gt,NX/2,NY/2));
  stml_multilayer_solve(p,"chip3",gt);
  printf("T33(inf)= %5.2f\n",stml_grid_read(gt,NX/2,NY/2));

  stml_grid_free(gq);
  stml_grid_free(gt);
  stml_grid_free(g0);
  stml_multilayer_free(p);
}
