/*
The first part of this example is the same as example_stack.c
In the second part, we use the temperature map computed by the
first part to recover the original power density map
*/

#include <stdio.h>
#include "stimul.h"

#define AMBIENT 30

// use grids of 160 x 160
#define NX 160
#define NY NX

#define H1 10
#define H2 500 


int main()
{
  int i,j;

  stml_stack *s = stml_stack_new();
  stml_grid *g = stml_grid_new(NX,NY);

  // set the two square heat sources in grid g
  stml_grid_set(g,0);
  stml_grid_rectangle(g,0,NX/2-NX/8,NX/4-1,NX/2+NX/8-1,1e6);
  stml_grid_rectangle(g,NX-1,NX/2-NX/8,NX-NX/4,NX/2+NX/8-1,1e6);

  // define 3 multilayers
  stml_stack_start(s,0.016,0.016,NX,NY,NULL,AMBIENT,"ML1");
  stml_stack_add(s,0.03,0.03,NX,NY,0,AMBIENT,"ML2");
  stml_stack_add(s,0.06,0.06,NX,NY,0,AMBIENT,"ML3");
  // first multilayer
  stml_stack_interface(s,"ML1",H1);
  stml_stack_slab(s,"ML1",1e-5,12);
  stml_stack_sources(s,"ML1",g,"circuits");
  stml_stack_slab(s,"ML1",5e-4,100);
  stml_stack_slab(s,"ML1",5e-5,5);
  // second multilayer
  stml_stack_slab(s,"ML2",1e-3,400);
  stml_stack_slab(s,"ML2",5e-5,5);
  // third multilayer
  stml_stack_slab(s,"ML3",5e-3,400);
  stml_stack_interface(s,"ML3",H2);
  // finalize and solve
  stml_stack_finalize(s,NULL,AMBIENT);
  stml_stack_solve(s,1e-3);
  // get temperature map of the source layer
  stml_stack_solution(s,"ML1","circuits",g);
  stml_stack_free(s);

  // Now we want to recover power density from temperature

  stml_grid *g1 = stml_grid_new(NX,NY);
  stml_grid *g2 = stml_grid_new(NX,NY);

  // compute heat flux going upward
  stml_multilayer *p = stml_multilayer_new(0.016,0.016,NX,NY);
  stml_multilayer_isotherm(p,AMBIENT,"top"); 
  stml_multilayer_interface(p,H1);
  stml_multilayer_slab(p,1e-5,12);
  stml_multilayer_temperature(p,g,"bottom"); 
  stml_multilayer_finalize(p);
  stml_multilayer_solve(p,"bottom",g1);

  // compute heat flux going downward
  s = stml_stack_new();
  stml_stack_start(s,0.016,0.016,NX,NY,g,0,"ML1");
  stml_stack_add(s,0.03,0.03,NX,NY,0,AMBIENT,"ML2");
  stml_stack_add(s,0.06,0.06,NX,NY,0,AMBIENT,"ML3");
  stml_stack_slab(s,"ML1",5e-4,100);
  stml_stack_slab(s,"ML1",5e-5,5);
  stml_stack_slab(s,"ML2",1e-3,400);
  stml_stack_slab(s,"ML2",5e-5,5);
  stml_stack_slab(s,"ML3",5e-3,400);
  stml_stack_interface(s,"ML3",H2);
  stml_stack_finalize(s,NULL,AMBIENT);
  stml_stack_solve(s,1e-3);
  stml_stack_solution(s,"ML1","top",g2);


  // Power density is the difference between downward and upward flux
  for (i=0; i<g->nx; i++) {
    for (j=0; j<g->ny; j++) {
      g->v[i][j] = g2->v[i][j] - g1->v[i][j];
    }
  }
  stml_grid_print(g);

  stml_grid_free(g);
  stml_grid_free(g1);
  stml_grid_free(g2);
  stml_stack_free(s);
}
