/*
  Copyright (C) 2008  Pierre Michaud
  Copyright (C) 2008  INRIA

  This file is part of STiMuL.

  STiMuL is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  STiMuL is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with STiMuL.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <fftw3.h>
#include "stimul.h"


#define ASSERT(cond) if (!(cond)) {fprintf(stderr,"assert line %d\n",__LINE__); exit(1);}
#define ASSERT2(cond,mess) if (!(cond)) {fprintf(stderr,mess); exit(1);}


#define IS_PASSIVE(lt) ((lt==stml_slab)||(lt==stml_interface))
#define IS_ACTIVE(lt) ((lt==stml_source)||(lt==stml_temperature))

#ifndef M_PI
#define M_PI (4*atan(1.))
#endif

#ifndef INFINITY
#define INFINITY HUGE_VAL
#endif

#define USE_FFT

#define FPACCERROR (1e-9)


/**********************************************************************************
           miscellaneous functions
 **********************************************************************************/


static void swapint(int *a, int *b)
{
  int c;
  c = *a;
  *a = *b;
  *b = c;
}


static void swapdouble(double *a, double *b)
{
  double c;
  c = *a;
  *a = *b;
  *b = c;
}


static double intersec(double x1, double x2, double xx1, double xx2)
{
  double y1,y2;
  if (x2<x1) swapdouble(&x1,&x2);
  if (xx2<xx1) swapdouble(&xx1,&xx2);
  y1 = (xx1 > x1)? xx1 : x1;
  y2 = (xx2 < x2)? xx2 : x2;
  if (y2>y1) {
    return y2-y1;
  } else {
    return 0;
  }
}




/**********************************************************************************
           string functions
 **********************************************************************************/


static void ss_init(stml_ss *p)
{
  int i;
  p->n = 0;
  for (i=0; i<STML_SSN; i++) {
    p->s[i][0] = '\0';
  }
}


static int ss_search(stml_ss *p, const char *s)
{
  int i,n;
  n = -1;
  for (i=0; i<p->n; i++) {
    if (strncmp(s,&p->s[i][0],STML_SSL)==0) {
      n = i;
      break;
    }
  }
  return n;
}


static void ss_add(stml_ss *p, const char *s)
{
  ASSERT(p->n < STML_SSN);
  if (ss_search(p,s) !=- 1) {
    fprintf(stderr,"STiMuL: '%s' is already used\n",s);
    ASSERT(0);
  }
  strncpy(&p->s[p->n][0],s,STML_SSL);
  p->n++;
}



/**********************************************************************************
           Grid functions
 **********************************************************************************/


void stml_grid_reset(stml_grid *g, int nx, int ny)
{
  int i,j;
  ASSERT(g);
  ASSERT(nx <= STML_GRIDMAX);
  ASSERT(ny <= STML_GRIDMAX);
  g->nx = nx;
  g->ny = ny;
  for (i=0; i<nx; i++) {
    for (j=0; j<ny; j++) {
      g->v[i][j] = 0;
    }
  }
}


void stml_grid_set(stml_grid *g, double val)
{
  int i,j;
  ASSERT(g);
  for (i=0; i<g->nx; i++) {
    for (j=0; j<g->ny; j++) {
      g->v[i][j] = val;
    }
  }
}


stml_grid * stml_grid_new(int nx, int ny)
{
  stml_grid *g;
  ASSERT(nx <= STML_GRIDMAX);
  ASSERT(ny <= STML_GRIDMAX);
  g = (stml_grid *) malloc(sizeof(stml_grid));
  ASSERT(g);
  stml_grid_reset(g,nx,ny);
  return g;
}


void stml_grid_free(stml_grid *g)
{
  free(g);
}


static void stml_grid_write(stml_grid *g, int i, int j, double val)
{
  ASSERT(g);
  ASSERT((i>=0) && (i<g->nx));
  ASSERT((j>=0) && (j<g->ny));
  g->v[i][j] = val;
}


static void stml_grid_acc(stml_grid *g, int i, int j, double val)
{
  ASSERT(g);
  ASSERT((i>=0) && (i<g->nx));
  ASSERT((j>=0) && (j<g->ny));
  g->v[i][j] += val;
}


static void stml_grid_multacc(stml_grid *g, stml_grid *g1, stml_grid *g2, char sub)
{
  int n,m;
  ASSERT(g);
  ASSERT(g1);
  ASSERT(g2);
  ASSERT(g1->nx == g2->nx);
  ASSERT(g1->ny == g2->ny);
  ASSERT(g->nx == g1->nx);
  ASSERT(g->ny == g1->ny);
  if (!sub) {
    for (n=0; n<g->nx; n++) {
      for (m=0; m<g->ny; m++) {
	g->v[n][m] += g1->v[n][m] * g2->v[n][m];
      }
    }
  } else {
    for (n=0; n<g->nx; n++) {
      for (m=0; m<g->ny; m++) {
	g->v[n][m] -= g1->v[n][m] * g2->v[n][m];
      }
    }
  }
}


void stml_grid_rectangle(stml_grid *g, int i1, int j1, int i2, int j2, double val)
{
  int i,j;
  ASSERT(g);
  ASSERT((i1>=0) && (i1<g->nx));
  ASSERT((j1>=0) && (j1<g->ny));
  ASSERT((i2>=0) && (i2<g->nx));
  ASSERT((j2>=0) && (j2<g->ny));
  if (i1>i2) swapint(&i1,&i2);
  if (j1>j2) swapint(&j1,&j2);
  for (i=i1; i<=i2; i++) {
    for (j=j1; j<=j2; j++) {
      g->v[i][j] = val;
    }
  }
}


static double fperror(double x, int nx)
{
  ASSERT(x>(-FPACCERROR));
  ASSERT(x<(nx+FPACCERROR));
  if (x<0) x=0;
  if (x>nx) x=nx;
  return x;
}


double stml_grid_read(stml_grid *g, double x, double y)
{
  int i,j;
  double dx,dy,a,b,c,z1,z2,z3,z4;
  ASSERT(g);
  x = fperror(x,g->nx);
  y = fperror(y,g->ny);
  ASSERT2((x>=0) && (x<=g->nx),"STiMuL: coordinates exceed grid dimensions\n");
  ASSERT2((y>=0) && (y<=g->ny),"STiMuL: coordinates exceed grid dimensions\n");
  ASSERT(g->nx>=2);
  ASSERT(g->ny>=2);
  i = x;
  j = y;
  if (((x-i)<0.5) && (i>0)) i--;
  if (((y-j)<0.5) && (j>0)) j--;
  if (i==(g->nx-1)) i--;
  if (j==(g->ny-1)) j--;
  ASSERT(i>=0);
  ASSERT(i<(g->nx-1));
  ASSERT(j>=0);
  ASSERT(j<(g->ny-1));
  z1 = g->v[i+1][j+1];
  z2 = g->v[i][j+1];
  z3 = g->v[i][j];
  z4 = g->v[i+1][j];
  a = (z1+z4-z2-z3)*0.5;
  b = (z1+z2-z3-z4)*0.5;
  c = (z1+z2+z3+z4)*0.25;
  dx = x-(i+1);
  dy = y-(j+1);
  return a*dx + b*dy + c;
}


void stml_grid_output(stml_grid *g, int i1, int j1, int i2, int j2, double step)
{
  double i,j;
  for (j=j2; j>=j1; j-=step) {
    for (i=i1; i<=i2; i+=step) {
      printf("%.8e ",stml_grid_read(g,i,j));
    }
    printf("\n");
  }
}


void stml_grid_print(stml_grid *g)
{
  int i,j;
  for (j=g->ny-1; j>=0; j--) {
    for (i=0; i<g->nx; i++) {
      printf("%.8e ",g->v[i][j]);
    }
    printf("\n");
  }
}


static void stml_grid_extract(stml_grid *g, stml_grid *gg, double x1, double y1, double x2, double y2)
{
  int i,j;
  double xx,yy,ex,ey;
  ASSERT(g);
  ASSERT(gg);
  x1 = fperror(x1,gg->nx);
  y1 = fperror(y1,gg->ny);
  x2 = fperror(x2,gg->nx);
  y2 = fperror(y2,gg->ny);
  ASSERT((x1>=0) && (x1<=gg->nx));
  ASSERT((y1>=0) && (y1<=gg->ny));
  ASSERT((x2>=0) && (x2<=gg->nx));
  ASSERT((y2>=0) && (y2<=gg->ny));
  ASSERT(g->nx>0);
  ASSERT(g->ny>0);
  ex = (x2-x1)/g->nx;
  ey = (y2-y1)/g->ny;
  for (i=0; i<g->nx; i++) {
    xx = x1 + (i+0.5)*ex;
    yy = y1 + 0.5*ey;
    for (j=0; j<g->ny; j++) {
      g->v[i][j] = stml_grid_read(gg,xx,yy);
      yy += ey;
    }
  }
}


double stml_grid_average(stml_grid *g)
{
  int i,j;
  double s = 0;
  for (i=0; i<g->nx; i++) {
    for (j=0; j<g->ny; j++) {
      s += g->v[i][j];
    }
  }
  return s / (g->nx*g->ny);
}


double stml_grid_max(stml_grid *g)
{
  int i,j;
  double m = g->v[0][0];
  for (i=0; i<g->nx; i++) {
    for (j=0; j<g->ny; j++) {
      if (g->v[i][j] > m) {
	m = g->v[i][j];
      }
    }
  }
  return m;
}


double stml_grid_min(stml_grid *g)
{
  int i,j;
  double m = g->v[0][0];
  for (i=0; i<g->nx; i++) {
    for (j=0; j<g->ny; j++) {
      if (g->v[i][j] < m) {
	m = g->v[i][j];
      }
    }
  }
  return m;
}




/**********************************************************************************
           cosine transforms
 **********************************************************************************/



#ifndef USE_FFT

static void dct2(stml_grid *gi, stml_grid *go)
{
  int i,j,n,m;
  double x;
  ASSERT(gi->nx > 0);
  ASSERT(gi->ny > 0);
  ASSERT(gi->nx < STML_GRIDMAX);
  ASSERT(gi->ny < STML_GRIDMAX);
  go->nx = gi->nx;
  go->ny = gi->ny;
  for (n=0; n<gi->nx; n++) {
    for (m=0; m<gi->ny; m++) {
      go->v[n][m] = 0;
      for (i=0; i<gi->nx; i++) {
	x = cos(M_PI*n*(i+0.5)/gi->nx);
	for (j=0; j<gi->ny; j++) {
	  go->v[n][m] += gi->v[i][j]*cos(M_PI*m*(j+0.5)/gi->ny)*x; 
	}
      }
    }
  }
}


static int dd(int n) 
{
  return (n==0)? 1:2;
}


static void idct2(stml_grid *gi, stml_grid *go)
{
  int i,j,n,m;
  double x,val;
  ASSERT(gi->nx > 0);
  ASSERT(gi->ny > 0);
  ASSERT(gi->nx < STML_GRIDMAX);
  ASSERT(gi->ny < STML_GRIDMAX);
  go->nx = gi->nx;
  go->ny = gi->ny;
  for (i=0; i<gi->nx; i++) {
    for (j=0; j<gi->ny; j++) {
      val = 0;
      for (n=0; n<gi->nx; n++) {
	x = dd(n)*cos(M_PI*n*(i+0.5)/gi->nx);
	for (m=0; m<gi->ny; m++) {
	  val += gi->v[n][m]*dd(m)*cos(M_PI*m*(j+0.5)/gi->ny)*x;
	} 
      }
      val = val / (gi->nx*gi->ny);
      stml_grid_write(go,i,j,val);
    }
  }
}


#else


static void fftw_dct2(stml_grid *gi, stml_grid *go)
{
  int i,j;
  fftw_plan p;
  double *in,*out;
  ASSERT(gi);
  ASSERT(go);
  ASSERT(gi->nx > 0);
  ASSERT(gi->ny > 0);
  ASSERT(gi->nx <= STML_GRIDMAX);
  ASSERT(gi->ny <= STML_GRIDMAX);
  go->nx = gi->nx;
  go->ny = gi->ny;
  in = (double*) fftw_malloc(sizeof(double)*gi->nx*gi->ny);
  out = (double*) fftw_malloc(sizeof(double)*gi->nx*gi->ny);
  p = fftw_plan_r2r_2d(gi->nx,gi->ny,in,out,FFTW_REDFT10,FFTW_REDFT10,FFTW_ESTIMATE);
  for (i=0; i<gi->nx; i++) {
    for (j=0; j<gi->ny; j++) {
      in[gi->ny*i+j] = gi->v[i][j];
    }
  }
  fftw_execute(p);
  for (i=0; i<gi->nx; i++) {
    for (j=0; j<gi->ny; j++) {
      go->v[i][j] = out[gi->ny*i+j]/4;
    }
  }
  fftw_destroy_plan(p);
  fftw_free(in); 
  fftw_free(out);
}


static void fftw_idct2(stml_grid *gi, stml_grid *go)
{
  int i,j;
  fftw_plan p;
  double *in,*out;
  ASSERT(go);
  ASSERT(gi);
  ASSERT(gi->nx > 0);
  ASSERT(gi->ny > 0);
  ASSERT(gi->nx <= STML_GRIDMAX);
  ASSERT(gi->ny <= STML_GRIDMAX);
  go->nx = gi->nx;
  go->ny = gi->ny;
  in = (double*) fftw_malloc(sizeof(double)*gi->nx*gi->ny);
  out = (double*) fftw_malloc(sizeof(double)*gi->nx*gi->ny);
  p = fftw_plan_r2r_2d(gi->nx,gi->ny,in,out,FFTW_REDFT01,FFTW_REDFT01,FFTW_ESTIMATE);
  for (i=0; i<gi->nx; i++) {
    for (j=0; j<gi->ny; j++) {
      in[gi->ny*i+j] = gi->v[i][j];
    }
  }
  fftw_execute(p);
  for (i=0; i<gi->nx; i++) {
    for (j=0; j<gi->ny; j++) {
      stml_grid_write(go,i,j,out[gi->ny*i+j]/(gi->nx*gi->ny));
    }
  }
  fftw_destroy_plan(p);
  fftw_free(in); 
  fftw_free(out);
}


#endif


static double coeff(double w, int nn, int n)
{
  double x,s;
  static stml_realtype *memo = NULL;
  static char *comp = NULL;
  int i;
  if (!memo) {
    memo = (stml_realtype *) malloc(STML_GRIDMAX*(STML_GRIDMAX+1)*sizeof(stml_realtype));
    comp = (char *) malloc(STML_GRIDMAX*(STML_GRIDMAX+1)*sizeof(char));
    for (i=0; i<(STML_GRIDMAX*(STML_GRIDMAX+1)); i++) {
      comp[i] = 0;
    }
  }
  ASSERT(memo);
  ASSERT(comp);
  ASSERT(n >= 0);
  ASSERT(nn > 0);
  ASSERT(n < nn);
  ASSERT(w > 0);
  ASSERT(nn <= STML_GRIDMAX);
  ASSERT(n < STML_GRIDMAX);
  x = (2*w/M_PI);
  if (n==0) {
    x *= M_PI/(2*nn);
  } else {
    i = nn*STML_GRIDMAX + n;
    if (comp[i]) {
      s = memo[i];
    } else {
      s = sin(M_PI*n/(2*nn))/n;
      memo[i] = s;
      comp[i] = 1;
    }
    x *= s;
  }
  return x;
}


static void ctran(double wx, double wy, stml_grid *gi, stml_grid *go)
{
  int n,m;
  double cn;
  ASSERT(gi);
  ASSERT(go);
#ifdef USE_FFT
  fftw_dct2(gi,go);
#else
  dct2(gi,go);
#endif
  for (n=0; n<go->nx; n++) {
    cn = coeff(wx,go->nx,n);
    for (m=0; m<go->ny; m++) {
      go->v[n][m] *= cn * coeff(wy,go->ny,m);
    }
  }
}


static void ictran(double wx, double wy, stml_grid *gi, stml_grid *go)
{
  int n,m;
  double cn;
  ASSERT(gi);
  ASSERT(go);
  for (n=0; n<gi->nx; n++) {
    cn = coeff(wx,gi->nx,n);
    for (m=0; m<gi->ny; m++) {
      gi->v[n][m] = gi->v[n][m] / (cn*coeff(wy,gi->ny,m));
    }
  }
#ifdef USE_FFT
  fftw_idct2(gi,go);
#else
  idct2(gi,go);
#endif
}



/**********************************************************************************
               Multilayer solver
 **********************************************************************************/


static double gamma_nm(stml_multilayer *p, int n, int m)
{
  double xx, yy;
  ASSERT(p);
  xx = n * p->ilx;
  yy = m * p->ily;
  return M_PI * sqrt(xx*xx+yy*yy);
}


static void thermal_quadrupole(stml_passive_layer *p, int n, int m, double gnm)
{
  double e,ie;
  ASSERT(p);
  ASSERT(n>=0);
  ASSERT(m>=0);
  ASSERT(n<STML_GRIDMAX);
  ASSERT(m<STML_GRIDMAX);
  ASSERT(p->v > 0);
  if (p->ltype==stml_interface) {
    p->hi->v[n][m] = 1./(2*p->v);
    p->va->v[n][m] = 0;
    return;
  }
  ASSERT(p->ltype==stml_slab);
  ASSERT(p->lz > 0);
  if ((n==0) && (m==0)) {
    p->hi->v[n][m] = p->lz/(2*p->v); 
    p->va->v[n][m] = 0;
    return;
  }
  ASSERT(gnm > 0);
  e = exp(gnm*p->lz);
  ASSERT(e > 1);
  ie = 1./e;
  p->va->v[n][m] = p->v * gnm * 0.5*(e-ie);
  if (isinf(e)) {
    p->hi->v[n][m] = 1. / (p->v * gnm);
  } else {
    p->hi->v[n][m] = (1+(ie*ie)-2*ie) / (p->v*gnm*(1-ie*ie));
  }
  ASSERT(!isnan(p->va->v[n][m]));
  ASSERT(!isnan(p->hi->v[n][m]));
}


static double equiv_imped(double hi, double va, double ii)
{
  /* 
  hi = horizontal impedance
  va = vertical admittance
  ii = input impedance
  */
  double imp;
  if ((ii+hi)==0) {
    return hi;
  }
  if (va==0) {
    return 2*hi+ii;
  }
  imp = hi + 1./(va + 1./(hi+ii));
  ASSERT(!isnan(imp));
  return imp;
}


static void equivalent_impedance(stml_grid *goi, stml_grid *gii, stml_passive_layer *p, int n, int m)
{
  goi->v[n][m] = equiv_imped(p->hi->v[n][m],p->va->v[n][m],(gii)? gii->v[n][m]:INFINITY);
}


static stml_passive_layer * passive_new(int num, stml_layer_type ltype, double lz, double val, int nx, int ny)
{
  stml_passive_layer *p;
  ASSERT(IS_PASSIVE(ltype));
  p = (stml_passive_layer *) malloc(sizeof(stml_passive_layer));
  ASSERT(p);
  p->num = num;
  p->ltype = ltype;
  p->lz = lz;
  p->v = val;
  p->hi = stml_grid_new(nx,ny);
  p->va = stml_grid_new(nx,ny);
  return p;
}


static void passive_free(stml_passive_layer *p)
{
  stml_grid_free(p->hi);
  stml_grid_free(p->va);
  free(p);
}


static stml_active_layer *active_new(int num, stml_layer_type ltype, stml_grid *g)
{
  stml_active_layer *p;
  ASSERT(IS_ACTIVE(ltype));
  p = (stml_active_layer *) malloc(sizeof(stml_active_layer));
  ASSERT(p);
  p->ltype = ltype;
  p->num = num;
  p->g = stml_grid_new(g->nx,g->ny);
  *p->g = *g;
  return p;
}


static void active_free(stml_active_layer *p)
{
  stml_grid_free(p->g);
  free(p);
}

stml_multilayer * stml_multilayer_new(double lx, double ly, int nx, int ny)
{
  int i,j;
  stml_multilayer *p;
  ASSERT(lx > 0);
  ASSERT(ly > 0);
  p = (stml_multilayer *) malloc(sizeof(stml_multilayer));
  ASSERT(p);
  p->lx = lx;
  p->ly = ly;
  p->nl = 0;
  p->np = 0;
  p->na = 0;
  p->nx = nx;
  p->ny = ny;
  for (i=0; i<STML_LAYERMAX; i++) {
    p->l[i].num = -1;
    p->p[i] = NULL;
    p->a[i] = NULL;
    p->imp0[i] = NULL;
    p->imp1[i] = NULL;
    for (j=0; j<STML_LAYERMAX; j++) {
      p->cm[i][j] = NULL;
    }
  }
  ss_init(&p->as);
  p->solved = NULL;
  p->ilx = 1/lx;
  p->ily = 1/ly;
  p->R = 0;
  return p;
}


static void stml_multilayer_freegrids(stml_multilayer *p)
{
  int i;
  ASSERT(p);
  for (i=0; i<p->np; i++) {
    stml_grid_free(p->p[i]->hi);
    stml_grid_free(p->p[i]->va);
    stml_grid_free(p->imp0[i]);
    stml_grid_free(p->imp1[i]);
    p->p[i]->hi = NULL;
    p->p[i]->va = NULL;
    p->imp0[i] = NULL;
    p->imp1[i] = NULL;
  }
}


void stml_multilayer_free(stml_multilayer *p)
{
  int i,j;
  ASSERT(p);
  if (!p->solved) return;
  for (i=0; i<p->np; i++) passive_free(p->p[i]);
  for (i=0; i<p->na; i++) active_free(p->a[i]);
  for (i=0; i<p->np; i++) stml_grid_free(p->imp0[i]);
  for (i=0; i<p->np; i++) stml_grid_free(p->imp1[i]);
  for (i=0; i<p->na; i++) {
    for (j=0; j<p->na; j++) {
      stml_grid_free(p->cm[i][j]);
    }
  }
  stml_grid_free(p->solved);
  p->solved = NULL;
  free(p);
}


static void stml_multilayer_layer(stml_multilayer *p, int num, stml_layer_type ltype)
{
  ASSERT2(p->nl < STML_LAYERMAX,"STiMuL: number of layers exceeds STML_LAYERMAX\n");
  p->l[p->nl].num = num;
  p->l[p->nl].ltype = ltype;
  p->nl++;
}


static void stml_multilayer_passive(stml_multilayer *p, stml_layer_type ltype, double lz, double val)
{
  ASSERT(p);
  ASSERT(IS_PASSIVE(ltype));
  ASSERT(p->np < STML_LAYERMAX);
  ASSERT(!p->p[p->np]);
  p->p[p->np] = passive_new(p->nl,ltype,lz,val,p->nx,p->ny);
  stml_multilayer_layer(p,p->np,ltype);
  p->np++;
}


static void stml_multilayer_active(stml_multilayer *p, stml_layer_type ltype, stml_grid *g, const char *name)
{
  ASSERT(p);
  ASSERT(IS_ACTIVE(ltype));
  ASSERT(p->na < STML_LAYERMAX);
  ASSERT(!p->a[p->na]);
  ASSERT((g->nx==p->nx)&&(g->ny==p->ny));
  p->a[p->na] = active_new(p->nl,ltype,g);
  ASSERT(p->a[p->na]->g->nx == p->nx);
  ASSERT(p->a[p->na]->g->ny == p->ny);
  stml_multilayer_layer(p,p->na,ltype);
  p->na++;
  ss_add(&p->as,name);
}


void stml_multilayer_isotherm(stml_multilayer *p, double t, const char *name)
{
  stml_grid *g;
  ASSERT(p);
  g = stml_grid_new(p->nx,p->ny);
  g->v[0][0] = t * p->lx * p->ly;
  stml_multilayer_active(p,stml_temperature,g,name);
  stml_grid_free(g);
}


void stml_multilayer_temperature(stml_multilayer *p, stml_grid *g, const char *name)
{
  stml_grid *gt;
  ASSERT(p);
  ASSERT(g);
  ASSERT2((g->nx==p->nx)&&(g->ny==p->ny),"STiMuL: grid mismatch\n");
  gt = stml_grid_new(p->nx,p->ny);
  ctran(p->lx,p->ly,g,gt);
  stml_multilayer_active(p,stml_temperature,gt,name);
  stml_grid_free(gt);
}


void stml_multilayer_sources(stml_multilayer *p, stml_grid *g, const char *name)
{
  stml_grid *gq;
  ASSERT(p); 
  gq = stml_grid_new(p->nx,p->ny);
  if (g) {
    ASSERT2((g->nx==p->nx)&&(g->ny==p->ny),"STiMuL: grid mismatch\n");
    ctran(p->lx,p->ly,g,gq);
  } else {
    /* null source */
    stml_grid_set(gq,0);
  }
  stml_multilayer_active(p,stml_source,gq,name);
  stml_grid_free(gq);
}


void stml_multilayer_interface(stml_multilayer *p, double h)
{
  ASSERT(p);
  ASSERT2(h>0,"STiMuL: interface thermal conductance must be > 0\n");
  stml_multilayer_passive(p,stml_interface,0,h);
  p->R += 1./(h * p->lx * p->ly);
}


void stml_multilayer_slab(stml_multilayer *p, double lz, double k)
{
  ASSERT(p);
  ASSERT2(lz>0,"STiMuL: slab cannot be infinitely thin, define interface instead\n");
  ASSERT2(k>0,"STiMuL: thermal conductivity must be > 0\n");
  stml_multilayer_passive(p,stml_slab,lz,k);
  p->R += lz/(k * p->lx * p->ly);
}


static void stml_multilayer_check(stml_multilayer *p)
{
  int i;
  ASSERT2(p->nl>=3,"STiMuL: There must be at least 3 layers\n");
  ASSERT2(IS_ACTIVE(p->l[0].ltype),"STiMuL: first layer must define a boundary condition\n");
  ASSERT2(IS_ACTIVE(p->l[p->nl-1].ltype),"STiMuL: last layer must define a boundary condition\n");
  ASSERT2((p->l[0].ltype==stml_temperature)||(p->l[p->nl-1].ltype==stml_temperature),"STiMuL: there must be at least one temperature (Dirichlet) boundary condition\n");
  for (i=1; i<p->nl; i++) {
    ASSERT2(IS_PASSIVE(p->l[i].ltype)||IS_PASSIVE(p->l[i-1].ltype),"STiMuL: active layers must be separated by passive ones\n");
  }
  for (i=1; i<(p->nl-1); i++) {
    ASSERT2(p->l[i].ltype!=stml_temperature,"STiMuL: inner layers can only be slabs, interfaces, or power sources\n");
  }
  ASSERT(p->R>0);
}


static void stml_multilayer_quadrupoles(stml_multilayer *p)
{
  int i,n,m;
  double gnm;
  for (n=0; n<p->nx; n++) {
    for (m=0; m<p->ny; m++) {
      gnm = gamma_nm(p,n,m);
      for (i=0; i<p->np; i++) {
	thermal_quadrupole(p->p[i],n,m,gnm);
	ASSERT(p->p[i]->hi->v[n][m] >= 0);
	ASSERT(p->p[i]->va->v[n][m] >= 0);
      }
    }
  }
}


static void stml_multilayer_impedances(stml_multilayer *p)
{
  stml_grid *nullimpedance = NULL;
  int i,n,m;

  for (i=0; i<p->np; i++) {
    ASSERT(!p->imp0[i]);
    ASSERT(!p->imp1[i]);
    p->imp0[i] = stml_grid_new(p->nx,p->ny);
    p->imp1[i] = stml_grid_new(p->nx,p->ny);
  }

  nullimpedance = stml_grid_new(p->nx,p->ny);
  stml_grid_set(nullimpedance,0);

  /* compute imp0 */
  for (n=0; n<p->nx; n++) {
    for (m=0; m<p->ny; m++) {
      i = 0;
      switch (p->l[0].ltype) {
      case stml_temperature: /* start with a short circuit */
	equivalent_impedance(p->imp0[i],nullimpedance,p->p[i],n,m);
	break;
      case stml_source: /* start with an open circuit */
	equivalent_impedance(p->imp0[i],NULL,p->p[i],n,m);
	break;
      default: ASSERT(0);
      }
      for (i=1; i<p->np; i++) {
	equivalent_impedance(p->imp0[i],p->imp0[i-1],p->p[i],n,m);
      }
    }
  }

  /* compute imp1 */
  for (n=0; n<p->nx; n++) {
    for (m=0; m<p->ny; m++) {
      i = p->np-1;
      switch (p->l[p->nl-1].ltype) {
      case stml_temperature: /* start with a short circuit */
	equivalent_impedance(p->imp1[i],nullimpedance,p->p[i],n,m);
	break;
      case stml_source: /* start with an open circuit */
	equivalent_impedance(p->imp1[i],NULL,p->p[i],n,m);
	break;
      default: ASSERT(0);
      }
      for (i=p->np-2; i>=0; i--) {
	equivalent_impedance(p->imp1[i],p->imp1[i+1],p->p[i],n,m);
      }
    }
  }

  stml_grid_free(nullimpedance);
}


static void stml_multilayer_propagate(stml_multilayer *p, int i, int n, int m)
{
  int j;
  double t,tt,ie,r,q,qq;
  stml_passive_layer *pl;
  ASSERT(i < p->na);
  ASSERT(p->cm[i][i]);

  q = t = INFINITY;

  if (p->a[i]->ltype==stml_source) {
    t = p->cm[i][i]->v[n][m];
    if (p->a[i]->num < (p->nl-1)) {
      ASSERT(IS_PASSIVE(p->l[p->a[i]->num+1].ltype));
      q = t / p->imp1[p->l[p->a[i]->num+1].num]->v[n][m];
    } else {
      q = 1;
    }
  } else {
    ASSERT(p->a[i]->ltype==stml_temperature);
    q = p->cm[i][i]->v[n][m];
    t = 1;
  }

  ASSERT(!isnan(q));
  ASSERT(!isnan(t));

  for (j=p->a[i]->num+1; j<p->nl; j++) {
    if (IS_PASSIVE(p->l[j].ltype)) {
      pl = p->p[p->l[j].num];
      if (p->l[j].num < (p->np-1)) {
	ie = p->imp1[p->l[j].num+1]->v[n][m];
      } else if (p->l[p->nl-1].ltype==stml_source) {
	ie = INFINITY;
      } else {
	ASSERT(p->l[p->nl-1].ltype==stml_temperature);
	ie = 0;
      }
      if ((pl->va->v[n][m]==0) && isinf(ie)) {
	r = 1;
      } else {
	r = 1. / (1 + pl->va->v[n][m] * (pl->hi->v[n][m]+ie));
      }
      qq = q * r;
      tt = t - pl->hi->v[n][m] * (q+qq);
      ASSERT(!isnan(qq));
      ASSERT(!isnan(tt));
    } else {
      ASSERT(IS_ACTIVE(p->l[j].ltype));
      ASSERT(p->cm[i][p->l[j].num]);
      if (p->l[j].ltype==stml_source) {
	ASSERT(!isnan(t));
	p->cm[i][p->l[j].num]->v[n][m] = tt = t;
	qq = q;
      } else {
	ASSERT(p->l[j].ltype==stml_temperature);
	ASSERT(j==(p->nl-1));
	ASSERT(!isnan(q));
	p->cm[i][p->l[j].num]->v[n][m] = qq = q;
	tt = t;
      }
    }
    t = tt;
    q = qq;
  }

}


static void stml_multilayer_contribution(stml_multilayer *p, int i)
{
  int n,m;
  stml_grid *i0,*i1;
  double imp;

  ASSERT(p);
  ASSERT(p->a[i]);
  ASSERT(i < p->na);

  /* compute (q,t) self-contribution for active layer i */
  ASSERT(p->cm[i][i]);
  i0 = i1 = NULL;
  if (p->a[i]->num > 0) {
    ASSERT(IS_PASSIVE(p->l[p->a[i]->num-1].ltype));
    i0 = p->imp0[p->l[p->a[i]->num-1].num];
  }
  if (p->a[i]->num < (p->nl-1)) {
    ASSERT(IS_PASSIVE(p->l[p->a[i]->num+1].ltype));
    i1 = p->imp1[p->l[p->a[i]->num+1].num];
  }
  switch (p->a[i]->ltype) {
  case stml_temperature:
    ASSERT(!i0 || !i1);
    for (n=0; n<p->nx; n++) {
      for (m=0; m<p->ny; m++) {
	if (i0) {
	  p->cm[i][i]->v[n][m] = -1. / i0->v[n][m];
	} else {
	  ASSERT(i1);
	  p->cm[i][i]->v[n][m] = 1. / i1->v[n][m];
	}
      }
    }
    break;
  case stml_source:
    ASSERT(i0 || i1);
    for (n=0; n<p->nx; n++) {
      for (m=0; m<p->ny; m++) {
	if (!i0) {
	  imp = i1->v[n][m];
	} else if (!i1) {
	  imp = i0->v[n][m];
	} else {
	  if ((i0->v[n][m]==0) || (i1->v[n][m]==0)) {
	    imp = 0;
	  } else {
	    imp = 1. / ((1./i0->v[n][m]) + (1./i1->v[n][m]));
	  }
	}
	p->cm[i][i]->v[n][m] = imp;
      }
    }
    break;
  default: ASSERT(0);
  }

  /* compute contributions on other active layers */
  for (n=0; n<p->nx; n++) {
    for (m=0; m<p->ny; m++) { 
      stml_multilayer_propagate(p,i,n,m);
    }
  }

}


static void stml_multilayer_print(stml_multilayer *p)
{
  int i;
  for (i=0; i<p->nl; i++) {
    switch (p->l[i].ltype) {
    case stml_temperature:
      fprintf(stderr,"[active] temperature '%s'\n",&p->as.s[p->l[i].num][0]); break;
    case stml_source:
      fprintf(stderr,"[active] source '%s'\n",&p->as.s[p->l[i].num][0]); break;
    case stml_slab:
      fprintf(stderr,"[passive] slab (Lz=%e,k=%e)\n",p->p[p->l[i].num]->lz,p->p[p->l[i].num]->v); break;
    case stml_interface:
      fprintf(stderr,"[passive] interface (h=%e)\n",p->p[p->l[i].num]->v); break;
    default: ASSERT(0);
    }
  }
}


void stml_multilayer_finalize(stml_multilayer *p)
{
  int i,j;
  ASSERT(p);
  stml_multilayer_print(p);
  stml_multilayer_check(p);
  stml_multilayer_quadrupoles(p);
  stml_multilayer_impedances(p);
  /* compute contribution matrix */
  /* (principle of superposition: consider each active layer separately, then sum contributions) */
  for (i=0; i<p->na; i++) {
    for (j=0; j<p->na; j++)  {
      ASSERT(!p->cm[i][j]);
      if (i <= j) { /* use networks reciprocity theorems */
	p->cm[i][j] = stml_grid_new(p->nx,p->ny);
      }
    }
  }
  for (i=0; i<p->na; i++) {
    stml_multilayer_contribution(p,i);
  }
  stml_multilayer_freegrids(p);
  ASSERT(!p->solved);
  p->solved = stml_grid_new(p->nx,p->ny);
}


static int active_layer_num(stml_multilayer *p, const char *name)
{
  int al = ss_search(&p->as,name);
  if (al<0) {
    fprintf(stderr,"STiMul: active layer name '%s' is unknown\n",name);
    ASSERT(0);
  }
  ASSERT(al<p->na);
  return al;
}


static void multilayer_modify(stml_multilayer *p, int al, stml_grid *g)
{
  ASSERT((al>=0) && (al<p->na));
  ASSERT(p->a[al]->g);
  ctran(p->lx,p->ly,g,p->a[al]->g);
}


void stml_multilayer_modify(stml_multilayer *p, const char *name, stml_grid *g)
{
  ASSERT(p);
  ASSERT(g);
  ASSERT2((p->nx==g->nx) && (p->ny==g->ny),"STiMuL: grid mismatch\n");
  int al = active_layer_num(p,name);
  multilayer_modify(p,al,g);
}


static void multilayer_solve(stml_multilayer *p, int al, stml_grid *g)
{
  int i;
  stml_grid *gg;
  char inv;
  ASSERT((al>=0) && (al<p->na));

  stml_grid_set(p->solved,0);

  /* sum contributions */
  for (i=0; i<p->na; i++) {
    if (i <= al) {
      gg = p->cm[i][al];
      inv = 0;
    } else {
      gg = p->cm[al][i];
      inv = (p->a[al]->ltype==stml_temperature) && (p->a[i]->ltype==stml_temperature);
    }
    stml_grid_multacc(p->solved,gg,p->a[i]->g,inv);
  }

  /* inverse transform */
  ictran(p->lx,p->ly,p->solved,g);
}


void stml_multilayer_solve(stml_multilayer *p, const char *name, stml_grid *g)
{
  ASSERT(p);
  ASSERT(g);
  ASSERT2(p->na>0,"STiMuL: no active layer defined\n");
  ASSERT2(p->solved,"STiMuL: multilayer not finalized\n");
  int al = active_layer_num(p,name);
  multilayer_solve(p,al,g);
}


/**********************************************************************************
               Multilayer-stack solver
 **********************************************************************************/


stml_stack * stml_stack_new()
{
  int i;
  stml_stack *s;
  s = (stml_stack *) malloc(sizeof(stml_stack));
  ASSERT(s);
  s->nm = 0;
  for (i=0; i<STML_MLMAX; i++) {
    s->m[i] = NULL;
    s->h[i] = INFINITY;
    s->t[i] = 0;
    s->gt[i] = NULL;
    s->ggt[i] = NULL;
    s->alpha[i] = 1;
  }
  ss_init(&s->ms);
  s->solved = 0;
  s->avgt[0] = s->avgt[1] = 0;
  return s;
}


void stml_stack_free(stml_stack *s)
{
  int i;
  ASSERT(s);
  for (i=0; i<s->nm; i++) {
    stml_multilayer_free(s->m[i]);
    stml_grid_free(s->gt[i]);
    stml_grid_free(s->ggt[i]);
    s->gt[i] = NULL;
    s->ggt[i] = NULL;
  }
  free(s);
}


void stml_stack_start(stml_stack *s, double lx, double ly, int nx, int ny, stml_grid *g, double t, const char *name)
{
  ASSERT(s);
  ASSERT2(s->nm==0,"STiMuL: stack cannot be started twice\n");
  s->m[0] = stml_multilayer_new(lx,ly,nx,ny);
  s->h[0] = 0;
  s->t[0] = 0;
  s->nm++;
  if (g) {
    stml_multilayer_temperature(s->m[0],g,STML_NAME0);
    s->avgt[0] = stml_grid_average(g);
  } else {
    stml_multilayer_isotherm(s->m[0],t,STML_NAME0);
    s->avgt[0] = t;
  }
  ss_add(&s->ms,name);
}


void stml_stack_add(stml_stack *s, double lx, double ly, int nx, int ny, double h, double t, const char *name)
{
  ASSERT(s);
  ASSERT2(s->nm<STML_MLMAX,"STiMuL: number of multilayers exceeds STML_MLMAX\n");
  ASSERT2(s->nm>=1,"STiMuL: you must start the stack with stml_stack_start\n");
  ASSERT2((lx>=s->m[s->nm-1]->lx)&&(ly>=s->m[s->nm-1]->ly),"STiMuL: can only stack multilayers of increasing size\n");
  ASSERT2(h>=0,"STiMuL: heat transfer coefficient must not be negative \n");
  s->m[s->nm] = stml_multilayer_new(lx,ly,nx,ny);
  stml_multilayer_sources(s->m[s->nm],NULL,STML_NAME0);
  s->h[s->nm] = h;
  s->t[s->nm] = t;
  s->nm++;
  ss_add(&s->ms,name);
}


static int multilayer_num(stml_stack *s, const char *name)
{
  int n = ss_search(&s->ms,name);
  if (n<0) {
    fprintf(stderr,"STiMul: multilayer '%s' is unknown\n",name);
    ASSERT(0);
  }
  ASSERT(n<s->nm);
  return n;
}


void stml_stack_sources(stml_stack *s, const char *name1, stml_grid *g, const char *name2)
{
  ASSERT(s);
  int n = multilayer_num(s,name1);
  stml_multilayer_sources(s->m[n],g,name2);
}


void stml_stack_interface(stml_stack *s, const char *name, double h)
{
  ASSERT(s);
  int n = multilayer_num(s,name);
  stml_multilayer_interface(s->m[n],h);
}


void stml_stack_slab(stml_stack *s, const char *name, double lz, double k)
{
  ASSERT(s);
  int n = multilayer_num(s,name);
  stml_multilayer_slab(s->m[n],lz,k);
}


static void stack_reset_gt(stml_stack *s)
{
  int i;
  double r,r0,r1,t;
  ASSERT(s);
  r = 0;
  for (i=0; i<s->nm; i++) {
    r += s->m[i]->R;
  }
  r0 = 0;
  r1 = r;
  for (i=1; i<s->nm; i++) {
    r0 += s->m[i-1]->R;
    r1 -= s->m[i-1]->R;
    t = (s->avgt[0] * r1 + s->avgt[1] * r0) / r;
    stml_grid_set(s->gt[i],t);
    stml_grid_set(s->ggt[i],t);
  }
}


void stml_stack_finalize(stml_stack *s, stml_grid *g, double t)
{
  int i;
  ASSERT(s);
  ASSERT2(s->nm>=2,"STiMuL: stack must comprise at least 2 multilayers\n");

  for (i=1; i<s->nm; i++) {
    ASSERT(s->gt[i]==NULL);
    s->gt[i] = stml_grid_new(s->m[i]->nx,s->m[i]->ny);
    ASSERT(s->ggt[i]==NULL);
    s->ggt[i] = stml_grid_new(s->m[i]->nx,s->m[i]->ny);
  }

  /* finalize multilayers */
  for (i=0; i<s->nm-1; i++) {
    fprintf(stderr,"--------------------------------------------------------------------\n");
    if (i==0) {
      fprintf(stderr,"MULTILAYER '%s': %.2e x %.2e\n",&s->ms.s[i][0],s->m[i]->lx,s->m[i]->ly);
    } else {
      fprintf(stderr,"MULTILAYER '%s': %.2e x %.2e  (h=%e,t=%e)\n",&s->ms.s[i][0],s->m[i]->lx,s->m[i]->ly,s->h[i],s->t[i]);
    }
    stml_multilayer_isotherm(s->m[i],s->t[i+1],STML_NAME1);
    stml_multilayer_finalize(s->m[i]);
  }
  i = s->nm-1;
  if (g) {
    stml_multilayer_temperature(s->m[i],g,STML_NAME1);
    s->avgt[1] = stml_grid_average(g);
  } else {
    stml_multilayer_isotherm(s->m[i],t,STML_NAME1);
    s->avgt[1] = t;
  }
    fprintf(stderr,"--------------------------------------------------------------------\n");
  fprintf(stderr,"MULTILAYER '%s': %.2e x %.2e  (h=%e,t=%e)\n",&s->ms.s[i][0],s->m[i]->lx,s->m[i]->ly,s->h[i],s->t[i]);
  stml_multilayer_finalize(s->m[i]);
  fprintf(stderr,"--------------------------------------------------------------------\n");
}


void stml_stack_modify(stml_stack *s, const char *name1, const char *name2, stml_grid *g)
{
  ASSERT(s);
  int n = multilayer_num(s,name1);
  stml_multilayer_modify(s->m[n],name2,g);
  if ((n==0) && (strcmp(name2,STML_NAME0)==0)) {
    s->avgt[0] = stml_grid_average(g);
  }
  if ((n==(s->nm-1)) && (strcmp(name2,STML_NAME1)==0)) {
    s->avgt[1] = stml_grid_average(g);
  }
}




static void stack_rect(stml_stack *s, int n, double *x1, double *y1, double *x2, double *y2)
{
  stml_multilayer *m,*p;
  ASSERT(n>=1);  
  m = s->m[n-1];
  p = s->m[n];
  *x1 = 0.5 * p->nx * (1 - m->lx*p->ilx);
  *y1 = 0.5 * p->ny * (1 - m->ly*p->ily);
  *x2 = 0.5 * p->nx * (1 + m->lx*p->ilx);
  *y2 = 0.5 * p->ny * (1 + m->ly*p->ily);
  ASSERT((*x2) > (*x1));
  ASSERT((*y2) > (*y1));
}



static void set_bc1(stml_stack *s, int n)
{
  double x1,x2,y1,y2;
  stml_grid *g;
  ASSERT(n<(s->nm-1));
  g = stml_grid_new(s->m[n]->nx,s->m[n]->ny);
  stack_rect(s,n+1,&x1,&y1,&x2,&y2);
  stml_grid_extract(g,s->gt[n+1],x1,y1,x2,y2);
  multilayer_modify(s->m[n],s->m[n]->na-1,g);
  stml_grid_free(g);
}


static void set_bc0(stml_stack *s, int n, stml_grid *gq)
{
  int i,j,ii,jj;
  stml_grid *g;
  double xx0,yy0,ex,ey,x00,y00,x10,y01,v,x1,y1,x2,y2;
  ASSERT(n>=1);
  ASSERT((gq->nx==s->m[n-1]->nx) && (gq->ny==s->m[n-1]->ny));
  g = stml_grid_new(s->m[n]->nx,s->m[n]->ny);

  xx0 = s->m[n]->nx * 0.5 * (1 - s->m[n-1]->lx * s->m[n]->ilx);
  yy0 = s->m[n]->ny * 0.5 * (1 - s->m[n-1]->ly * s->m[n]->ily);
  ex = s->m[n-1]->lx * s->m[n]->ilx * s->m[n]->nx / s->m[n-1]->nx;
  ey = s->m[n-1]->ly * s->m[n]->ily * s->m[n]->ny / s->m[n-1]->ny;

  /* contribution from previous multilayer */
  for (i=0; i<s->m[n-1]->nx; i++) {
    x00 = xx0 + i*ex;
    y00 = yy0;
    for (j=0; j<s->m[n-1]->ny; j++) {
      x10 = x00 + ex;
      y01 = y00 + ey;
      for (ii=(int)x00; ii<x10; ii++) {
	for (jj=(int)y00; jj<y01; jj++) {
	  v = gq->v[i][j] * intersec(ii,ii+1,x00,x10) * intersec(jj,jj+1,y00,y01);
	  stml_grid_acc(g,ii,jj,v);
	}
      }
      y00 += ey;
    }
  }

  /* contribution from exposed surface */
  stack_rect(s,n,&x1,&y1,&x2,&y2);
  for (ii=0; ii<g->nx; ii++) {
    for (jj=0; jj<g->ny; jj++) {
      v = (s->t[n] - s->gt[n]->v[ii][jj]) * s->h[n]; 
      v *= 1 - intersec(ii,ii+1,x1,x2) * intersec(jj,jj+1,y1,y2);
      stml_grid_acc(g,ii,jj,v);
    }
  }
  multilayer_modify(s->m[n],0,g);
  stml_grid_free(g);
}


static void update_t(stml_stack *s, int n, double *maxd, double *mt)
{
  int i,j;
  double d,u1,u2;
  ASSERT(n>=1);
  ASSERT(s->alpha[n]>0);
  *mt = 0;
  for (i=0; i<s->gt[n]->nx; i++) {
    for (j=0; j<s->gt[n]->ny; j++) {
      if (maxd) {
	d = fabs(s->ggt[n]->v[i][j] - s->gt[n]->v[i][j]);
	u1 = fabs(s->gt[n]->v[i][j]);
	u2 = fabs(s->ggt[n]->v[i][j]);
	if ((u1==0) && (u2==0)) {
	  d = 1;
	} else if (u1!=0) {
	  d = d/u1;
	} else {
	  d = d/u2;
	}
	if (d > (*maxd)) *maxd = d;
      }
      s->gt[n]->v[i][j] = s->alpha[n]*s->ggt[n]->v[i][j] + (1-s->alpha[n])*s->gt[n]->v[i][j];
      if (fabs(s->gt[n]->v[i][j]) > *mt) *mt = fabs(s->gt[n]->v[i][j]);
    }
  }
}


static void stml_stack_iteration(stml_stack *s, double *maxd, stml_grid *gq, double mt[])
{
  int n;
  if (maxd) *maxd = 0;
  set_bc1(s,0);
  multilayer_solve(s->m[0],s->m[0]->na-1,gq);
  for (n=1; n<=s->nm-1; n++) {
    set_bc0(s,n,gq);
    if (n<(s->nm-1)) set_bc1(s,n); 
    stml_grid_reset(gq,s->m[n]->nx,s->m[n]->ny);
    multilayer_solve(s->m[n],0,s->ggt[n]);
    if (n<(s->nm-1)) multilayer_solve(s->m[n],s->m[n]->na-1,gq);
    update_t(s,n,maxd,&mt[n]);
  }
}


#define ALPHARATIO1 0.95
#define ALPHARATIO2 0.5
#define ITERFACTOR 1000
#define MAXITER 200
#define NMD 10
#define UNKNOWN (-1)
#define NTRIES 2
#define MINALPHA (1e-5)


static void update_h(double h[],double md, int n)
{
  int i;
  for (i=n-1; i>=1; i--) {
    h[i] = h[i-1];
  }
  h[0] = md;
}


static int expected_time(double h[], int n, double eps)
{
  int i;
  ASSERT(n>=2);
  if (h[0] <= eps) return 0;
  for (i=1; i<n; i++) {
    if (h[i] <= h[i-1]) return UNKNOWN;
  }
  ASSERT(h[1] > eps);
  ASSERT(h[1] > h[0]);
  return (log(eps)-log(h[1])) / (log(h[0])-log(h[1]));
}



void stml_stack_solve(stml_stack *s, double eps)
{
  int i,n,nn,et;
  char diverge,noconverge;
  double md;
  double hmd[NMD];
  int tries = 0;
  stml_grid *gq = stml_grid_new(s->m[0]->nx,s->m[0]->ny);
  double mt[STML_MLMAX];

  ASSERT(s);
  stack_reset_gt(s);
  for (i=1; i<s->nm; i++) s->alpha[i]=1;
  n = nn = 0;
  et = UNKNOWN;
  for (i=0; i<NMD; i++) {
    hmd[i]=1;
  }

  while (1) {
    n++; 
    nn++;
    stml_stack_iteration(s,&md,gq,mt);
    fprintf(stderr,"%d : %e\n",n,md);
    if (md <= eps) {
      break;
    }
    update_h(hmd,md,NMD);
    if (nn >= NMD) {
      et = expected_time(hmd,NMD,eps);
      if (et>=0) {
	fprintf(stderr,"Expected convergence time: %d\n",n+et);
      }
    }
    diverge = 0;
    for (i=1; i<s->nm; i++) {
      if (mt[i] > STML_MAXTEMP) {
	diverge = 1;
	break;
      }
    }
    noconverge = (nn>MAXITER) && ((et==UNKNOWN) || (et > (nn*ITERFACTOR)));
    if (diverge) {
      fprintf(stderr,"DIVERGE\n");
      for (i=1; i<s->nm; i++) {
	if (mt[i] > STML_MAXTEMP) {
	  s->alpha[i] = s->alpha[i] * ALPHARATIO2;
	} else {
	  s->alpha[i] = s->alpha[i] * ALPHARATIO1;
	}
	ASSERT2(s->alpha[i]>MINALPHA,"STiMuL: convergence problem. You may try to increase the value of STML_MAXTEMP\n");
      }
      stack_reset_gt(s);
      nn = 0;
      tries = 0;
    } else if (noconverge && (tries<NTRIES)) {
      for (i=1; i<s->nm; i++) {
	s->alpha[i] = s->alpha[i] * ALPHARATIO1;
      }
      stack_reset_gt(s);
      nn = 0;
      tries++;
    }
    et = UNKNOWN;
  }
  s->solved = 1;
  stml_grid_free(gq);
}


void stml_stack_solution(stml_stack *s, const char *name1, const char *name2, stml_grid *g)
{
  ASSERT(s);
  ASSERT2(s->solved,"STiMuL: stack is not solved\n");
  int n = multilayer_num(s,name1);
  stml_multilayer_solve(s->m[n],name2,g);
}

