/*
  Copyright (C) 2008  Pierre Michaud
  Copyright (C) 2008  INRIA

  This file is part of STiMuL.

  STiMuL is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  STiMuL is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with STiMuL.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef STIMUL_H
#define STIMUL_H

#define STML_GRIDMAX 2000
#define STML_MAXTEMP 1e5
#define STML_LAYERMAX 20
#define STML_MLMAX 10
#define STML_SSN 20
#define STML_SSL 60
#define STML_NAME0 "top"
#define STML_NAME1 "bottom"


typedef double stml_realtype;


typedef struct {
  int nx;
  int ny;
  stml_realtype v[STML_GRIDMAX][STML_GRIDMAX];
} stml_grid;


typedef struct {
  char s[STML_SSN][STML_SSL+1];
  int n;
} stml_ss;


typedef enum {stml_temperature,stml_source,stml_slab,stml_interface} stml_layer_type;


typedef struct {
  stml_layer_type ltype;
  int num;
  double lz;
  double v;
  /* thermal quadrupole (aka two-port network) */
  stml_grid *hi; /* horizontal impedances */
  stml_grid *va; /* vertical admittance */
} stml_passive_layer;


typedef struct {
  stml_layer_type ltype;
  int num;
  stml_grid *g;
} stml_active_layer;


typedef struct {
  stml_layer_type ltype;
  int num;
} stml_layer;


typedef struct {
  double lx;
  double ly;
  stml_layer l[STML_LAYERMAX];
  stml_passive_layer *p[STML_LAYERMAX];
  stml_active_layer *a[STML_LAYERMAX];
  stml_ss as;
  int nl;
  int np;
  int na;
  int nx;
  int ny;
  double ilx;
  double ily;
  stml_grid *imp0[STML_LAYERMAX]; /* equivalent impedance layers [0,l] */
  stml_grid *imp1[STML_LAYERMAX]; /* equivalent impedance layers [l,np-1] */
  stml_grid *cm[STML_LAYERMAX][STML_LAYERMAX]; /* contribution matrix */
  stml_grid *solved;
  double R;
} stml_multilayer;


typedef struct {
  int nm;
  stml_multilayer *m[STML_MLMAX];
  stml_ss ms;
  double h[STML_MLMAX];
  double t[STML_MLMAX];
  stml_grid *gt[STML_MLMAX];
  stml_grid *ggt[STML_MLMAX];
  char solved;
  double alpha[STML_MLMAX];
  double avgt[2];
} stml_stack;


stml_grid * stml_grid_new(int,int);
void stml_grid_free(stml_grid *);
void stml_grid_reset(stml_grid *,int,int);
void stml_grid_set(stml_grid *,double);
void stml_grid_rectangle(stml_grid *,int,int,int,int,double);
double stml_grid_read(stml_grid *,double,double);
void stml_grid_output(stml_grid *,int,int,int,int,double);
void stml_grid_print(stml_grid *);
double stml_grid_average(stml_grid *);
double stml_grid_max(stml_grid *);
double stml_grid_min(stml_grid *);

stml_multilayer * stml_multilayer_new(double,double,int,int);
void stml_multilayer_free(stml_multilayer *);
void stml_multilayer_isotherm(stml_multilayer *,double,const char *);
void stml_multilayer_temperature(stml_multilayer *,stml_grid *,const char *);
void stml_multilayer_sources(stml_multilayer *,stml_grid *,const char *);
void stml_multilayer_interface(stml_multilayer *,double);
void stml_multilayer_slab(stml_multilayer *,double,double);
void stml_multilayer_finalize(stml_multilayer *);
void stml_multilayer_modify(stml_multilayer *,const char *,stml_grid *);
void stml_multilayer_solve(stml_multilayer *,const char *,stml_grid *);

stml_stack * stml_stack_new();
void stml_stack_free(stml_stack *);
void stml_stack_start(stml_stack *,double,double,int,int,stml_grid *,double,const char *);
void stml_stack_add(stml_stack *,double,double,int,int,double,double,const char *);
void stml_stack_sources(stml_stack *,const char *,stml_grid *,const char *);
void stml_stack_interface(stml_stack *,const char *,double);
void stml_stack_slab(stml_stack *,const char *,double,double);
void stml_stack_finalize(stml_stack *,stml_grid *,double);
void stml_stack_modify(stml_stack *,const char *,const char *,stml_grid *);
void stml_stack_solve(stml_stack *,double);
void stml_stack_solution(stml_stack *,const char *,const char *,stml_grid *);


#endif
